# -*- coding: utf-8 -*-
"""
/***************************************************************************
 Sample
                                 A QGIS plugin
 QGIS Sample Plugin
                              -------------------
        begin                : 2020-06-01
        git sha              : $Format:%H$
        copyright            : (C) 2020 by MIERUNE Inc.
        email                : info@mierune.co.jp
        license              : GNU General Public License v2.0
 ***************************************************************************/
"""

import os
import csv
import math

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from qgis.core import *
from qgis.gui import *

from .Tiff2PngBase import Tiff2PngBase
from .GeoTiffManager import GeoTiffManager
from .PngProcessor import PngProcessor
from .RubberBandMapTool import RubberbandMapTool
from .. import settings


class Tiff2Png:
    def __init__(self, iface):
        self.iface = iface
        self.canvas = self.iface.mapCanvas()

        self.dlg = Tiff2PngBase()
        self.init_gui()

        # ボタン設定
        self.dlg.pushButton_export_png.clicked.connect(self.area_export)
        self.dlg.pushButton_run.clicked.connect(self.enable_area_selection_mode)
        self.dlg.pushButton_runCenter.clicked.connect(self.enable_area_selection_center_mode)
        self.dlg.pushButton_cancel.clicked.connect(self.dlg.close)
        self.dlg.bUpperMarginAuto.toggled.connect(self.UpperMarginAutoChange)

    # すべてのUIパーツを初期化
    def init_gui(self):
        # 1ピクセルあたりの地理的距離スピンボックスの小数点以下桁数をセット
        self.dlg.outputDistXSpinBox.setDecimals(
            settings.PX_DIST_DECIMAL_DIGITS)
        self.dlg.outputDistYSpinBox.setDecimals(
            settings.PX_DIST_DECIMAL_DIGITS)
        self.dlg.lowerMarginSpinBox.setDecimals(
            settings.HEIGHT_MARGIN_DECIMAL_DIGITS)
        self.dlg.upperMarginSpinBox.setDecimals(
            settings.HEIGHT_DECIMAL_DIGITS)

        # UIのイベントに処理を紐付け
        self.dlg.demSelectionComboBox.setFilters(
            QgsMapLayerProxyModel.RasterLayer)
        self.dlg.demSelectionComboBox.currentIndexChanged.connect(
            self.init_labels_buttons_geoinfo)
        self.dlg.outputFileWidget.fileChanged.connect(self.init_labels_buttons)
        self.dlg.outputFileWidget.setStorageMode(QgsFileWidget.GetDirectory)
        self.dlg.upperMarginSpinBox.valueChanged.connect(
            self.recalc_upper_margin)
        self.dlg.lowerMarginSpinBox.valueChanged.connect(
            self.recalc_upper_margin)
        self.dlg.demSelectionComboBox.currentIndexChanged.connect(
            self.recalc_upper_margin)

        self.dlg.AreaSelectionComboBox.setFilters(
            QgsMapLayerProxyModel.RasterLayer)
        self.dlg.AreaSelectionComboBox.currentIndexChanged.connect(
            self.init_area_selection)
        # self.dlg.inputLayerOriginXStaSpinBox
        # self.dlg.inputLayerOriginYStaSpinBox
        # self.dlg.inputLayerOriginXEndSpinBox
        # self.dlg.inputLayerOriginYEndSpinBox

        # その他UIパーツの初期化
        self.init_labels_buttons_geoinfo()
        self.init_resolutions_combobox()
        self.recalc_upper_margin()

    def init_labels_buttons(self):
        rlayer = self.dlg.demSelectionComboBox.currentLayer()

        # UIの入力状態が不正ならエラー表示して実行ボタンの無効化
        # エラー表示を初期化
        self.dlg.inputErrorMessageLabel.setText("")
        self.dlg.outputErrorMessageLabel.setText("")
        # 実行可かどうかの判定結果
        is_executable = True
        # 非対応レイヤーが選択されている場合
        if self.is_invalid(rlayer):
            self.dlg.inputErrorMessageLabel.setText(self.is_invalid(rlayer))
            is_executable = False
        # 出力先が指定されていない場合
        if self.dlg.outputFileWidget.filePath() == "":
            self.dlg.outputErrorMessageLabel.setText("ファイルの出力先を指定して下さい")
            is_executable = False
        # 標高幅が大きすぎて上方マージンが負の値となる場合
        elif self.dlg.upperMarginSpinBox.value() < 0:
            self.dlg.outputErrorMessageLabel.setText("標高幅が規定値を超過しています")
            is_executable = False
        # 問題なければ実行可に
        self.dlg.pushButton_run.setEnabled(is_executable)
        self.dlg.pushButton_runCenter.setEnabled(is_executable)

        if self.dlg.inputLayerOriginXStaSpinBox.value() == 0 and self.dlg.inputLayerOriginYStaSpinBox.value() == 0 and self.dlg.inputLayerOriginXEndSpinBox.value() == 0 and self.dlg.inputLayerOriginYEndSpinBox.value() == 0:
            is_executable = False
        self.dlg.pushButton_export_png.setEnabled(is_executable)

        # 有効なレイヤーがないエラーの場合
        # レイヤー情報のラベルをすべて空欄に
        if rlayer is None:
            self.dlg.inputResolutionLabel.setText("")
            self.dlg.inputGeoInfoLabel.setText("")
            # self.dlg.inputDistXLabel.setText("")
            # self.dlg.inputDistYLabel.setText("")
            self.dlg.inputCRSLabel.setText("")
            return

    def init_labels_buttons_geoinfo(self):
        self.init_labels_buttons()

        # QGIS-APIとGDALでGeoTiffの各種情報を取得
        rlayer = self.dlg.demSelectionComboBox.currentLayer()
        gt_manager = GeoTiffManager(rlayer.source())
        size = gt_manager.get_info("size")
        x_dist, y_dist = gt_manager.get_distance_per_pixel()
        tif_min, tif_max = gt_manager.get_tiff_min_max()
        crs_name = rlayer.crs().description()

        # 値をUIにセット
        self.dlg.inputResolutionLabel.setText(f"{size[0]}x{size[1]}")
        self.dlg.inputGeoInfoLabel.setText(
            f"x方向:{str(x_dist)}m " +
            f"y方向:{str(y_dist*-1)}m " + 
            f"標高最小:{str(round(tif_min, settings.HEIGHT_DECIMAL_DIGITS))}m " +
            f"標高最大:{str(round(tif_max, settings.HEIGHT_DECIMAL_DIGITS))}m")
        self.dlg.inputCRSLabel.setText(crs_name)
        self.dlg.outputDistXSpinBox.setValue(x_dist)
        self.dlg.outputDistYSpinBox.setValue(y_dist * -1)
        self.dlg.lowerMarginSpinBox.setValue(20)
        self.dlg.upperMarginSpinBox.setValue(20)
        # self.dlg.lowerMarginSpinBox.setValue(math.floor(tif_min))
        # self.dlg.upperMarginSpinBox.setValue(math.ceil(tif_max))

    def recalc_upper_margin(self):
        rlayer = self.dlg.demSelectionComboBox.currentLayer()
        # 非対応レイヤーが選択されている場合は計算しない
        if self.is_invalid(rlayer):
            return

        # GeoTIFFの情報取得
        gt_manager = GeoTiffManager(rlayer.source())
        tif_min, tif_max = gt_manager.get_tiff_min_max()
        lower_margin = round(self.dlg.lowerMarginSpinBox.value(), settings.HEIGHT_MARGIN_DECIMAL_DIGITS)
        if self.dlg.bUpperMarginAuto.isChecked():
            # 標高幅リストを小さい順に走査して、可能な限り小さい標高幅を用いて最高標高を算出する
            for diff in settings.HEIGHT_DIFF_LIST:
                src_max = tif_min - lower_margin + diff
                if (tif_max - tif_min + lower_margin) > diff:
                    continue
                else:
                    break

            # 上方マージンを算出してUIにセット
            upper_margin = round(src_max - tif_max, settings.HEIGHT_DECIMAL_DIGITS)
            self.dlg.upperMarginSpinBox.setValue(upper_margin)
        else:
            upper_margin = round(self.dlg.upperMarginSpinBox.value(), settings.HEIGHT_MARGIN_DECIMAL_DIGITS)

        src_min, src_max = gt_manager.calc_src_min_max(self.dlg.bUpperMarginAuto.isChecked(), lower_margin, upper_margin)
        src_dif = src_max - src_min
        self.dlg.HeightMapInfoLabel.setText(
            f"高さステップ:{round(src_dif/65535, 5)}m    " + 
            f"HeightMap最小値:{round((tif_min-src_min)*65535/src_dif, 0)}  " +
            f"HeightMap最大値:{round((tif_max-src_min)*65535/src_dif, 0)}")

    def init_area_selection(self):
        rlayer = self.dlg.AreaSelectionComboBox.currentLayer()
        # 非対応レイヤーが選択されている場合は計算しない
        if self.is_invalid(rlayer):
            return

        self.dlg.inputLayerOriginXStaSpinBox.setValue(rlayer.dataProvider().extent().xMinimum())
        self.dlg.inputLayerOriginYStaSpinBox.setValue(rlayer.dataProvider().extent().yMaximum())
        self.dlg.inputLayerOriginXEndSpinBox.setValue(rlayer.dataProvider().extent().xMaximum())
        self.dlg.inputLayerOriginYEndSpinBox.setValue(rlayer.dataProvider().extent().yMinimum())

        self.dlg.pushButton_export_png.setEnabled(True)


    def is_invalid(self, layer: QgsRasterLayer):
        """[summary]
        入力レイヤーが有効なレイヤーかどうかチェックする
        もし有効なレイヤーではない場合、エラーメッセージを返す
        有効ならNoneを返す

        Args:
            layer (QgsMapLayer): 入力レイヤー

        Returns:
            str: エラーメッセージ、Noneになりうる
        """
        if layer is None:
            return "プロジェクトに有効なレイヤーが存在しません"

        epsg_code = layer.crs().authid()
        if epsg_code == "":
            return "入力画像の座標系が不明です"

        if not int(epsg_code.split(":")[-1]) in settings.SUPPORTED_EPSG_CODES:
            return "入力画像は平面直角座標系でなければなりません"

        if not layer.bandCount() == 1:
            return "入力画像はグレースケールでなければなりません"

        return None

    def init_resolutions_combobox(self):
        for name, resolutions in settings.SUPPORTED_RESOLUTIONS.items():
            for res in resolutions:
                self.dlg.outputResolutionsComboBox.addItem(
                    f"{name}: {res}", res)

        # カスタム解像度の場合コンボボックスの値がNoneとする
        self.dlg.outputResolutionsComboBox.addItem("任意解像度(手動入力)", None)
        self.dlg.outputResolutionsComboBox.currentIndexChanged.connect(
            self.set_enabled_custom_resolution)

    def set_enabled_custom_resolution(self):
        # カスタム解像度のコンボボックスの値はNone
        custom_enabled = self.dlg.outputResolutionsComboBox.currentData() is None
        self.dlg.customWidthSpinBox.setEnabled(custom_enabled)
        self.dlg.customHeightSpinBox.setEnabled(custom_enabled)

    def UpperMarginAutoChange(self):
        self.dlg.upperMarginSpinBox.setEnabled(self.dlg.bUpperMarginAuto.isChecked() == False)
        self.recalc_upper_margin()

    def area_export(self):
        rlayer = self.dlg.AreaSelectionComboBox.currentLayer()
        if rlayer is not None:
            self.export_png([0, 0], [0, 0], 1, rlayer)
        else:
            QMessageBox.information(None, '範囲レイヤー指定', '範囲指定するレイヤーを選択して下さい')

    # 切り出し範囲選択ボタンクリック時
    def enable_area_selection_mode(self):
        # 処理前にプロジェクトCRSをDEMのCRSに合わせる
        QgsProject.instance().setCrs(self.dlg.demSelectionComboBox.currentLayer().crs())

        QMessageBox.information(None, '処理開始', 'キャンバス上で切り出し範囲を指定して下さい')
        self.dlg.hide()

        # 1ピクセルの距離とPNG解像度からタイル1枚の地理的サイズを計算
        x_dist = self.dlg.outputDistXSpinBox.value()
        y_dist = self.dlg.outputDistYSpinBox.value()
        output_resolution = self.get_output_resolution()
        output_space = self.get_output_space()
        tilesize = [(output_resolution[0] - output_space[0] * 2) * x_dist,
                    (output_resolution[1] - output_space[1] * 2) * y_dist]

        # ラバーバンドマップツールを有効化
        mt_extent = RubberbandMapTool(
            self.iface, self.on_area_selected, tilesize, False)
        self.iface.mapCanvas().setMapTool(mt_extent)

    # 切り出し範囲選択ボタンクリック時 (中心から)
    def enable_area_selection_center_mode(self):
        # 処理前にプロジェクトCRSをDEMのCRSに合わせる
        QgsProject.instance().setCrs(self.dlg.demSelectionComboBox.currentLayer().crs())

        QMessageBox.information(None, '処理開始', 'キャンバス上で切り出し範囲を指定して下さい(中心から)')
        self.dlg.hide()

        # 1ピクセルの距離とPNG解像度からタイル1枚の地理的サイズを計算
        x_dist = self.dlg.outputDistXSpinBox.value()
        y_dist = self.dlg.outputDistYSpinBox.value()
        output_resolution = self.get_output_resolution()
        output_space = self.get_output_space()
        tilesize = [(output_resolution[0] - output_space[0] * 2) * x_dist,
                    (output_resolution[1] - output_space[1] * 2) * y_dist]

        # ラバーバンドマップツールを有効化
        mt_extent = RubberbandMapTool(
            self.iface, self.on_area_selected, tilesize, True)
        self.iface.mapCanvas().setMapTool(mt_extent)

    # 切り出し範囲確定時のアクション
    def on_area_selected(self, lefttop_xy, rightbottom_xy, quadrant, grids_count):
        rlayer = self.dlg.demSelectionComboBox.currentLayer()
        if self.is_invalid(rlayer):
            # 通常発生しない
            print(self.is_invalid(rlayer))
            return

        # 生成予定のタイルが0枚の場合処理を中断
        if grids_count == 0:
            QMessageBox.information(
                None, 'エラー', '選択領域が小さすぎます\n選択領域はタイル1枚以上の大きさでなければなりません')
            self.dlg.show()
            return

        self.export_png(lefttop_xy, rightbottom_xy, quadrant)

    # PNG切り出し
    def export_png(self, lefttop_xy, rightbottom_xy, quadrant=1, select_layer=None):
        rlayer = self.dlg.demSelectionComboBox.currentLayer()
        lower_margin = self.dlg.lowerMarginSpinBox.value()
        upper_margin = self.dlg.upperMarginSpinBox.value()
        geotiff = GeoTiffManager(rlayer.source())
        x_dist = self.dlg.outputDistXSpinBox.value()
        y_dist = self.dlg.outputDistYSpinBox.value()
        authid = QgsProject.instance().crs().authid()

        if select_layer is not None:
            lefttop_xy = [self.dlg.inputLayerOriginXStaSpinBox.value(), self.dlg.inputLayerOriginYStaSpinBox.value()]
            rightbottom_xy = [self.dlg.inputLayerOriginXEndSpinBox.value(), self.dlg.inputLayerOriginYEndSpinBox.value()]
            authid = projWinSRS=select_layer.crs().authid()

        output_space = self.get_output_space()
        output_space[0] =  output_space[0] * x_dist
        output_space[1] = -output_space[1] * y_dist
        lefttop_xy = [x - y for x, y in zip(lefttop_xy, output_space)]
        rightbottom_xy = [x + y for x, y in zip(rightbottom_xy, output_space)]
        
        QMessageBox.information(None, "通知",
            "\nexport_png  lefttop_xy={}  rightbottom_xy={}  authid={}  quadrant={}\n".format(lefttop_xy, rightbottom_xy, authid, quadrant))

        png_data = geotiff.translate_to_png(
            bUpperMarginAuto=self.dlg.bUpperMarginAuto.isChecked(),
            lower_margin=lower_margin,
            upper_margin=upper_margin,
            projWin=lefttop_xy + rightbottom_xy,
            projWinSRS=authid,
            xRes=x_dist,
            yRes=y_dist)

        # 出力先ディレクトリにoutputフォルダを作成
        output_path = os.path.join(
            self.dlg.outputFileWidget.filePath(), "output")
        os.makedirs(output_path, exist_ok=True)

        output_resolution = self.get_output_resolution()

        # PNG出力
        png_path = png_data.get("path")
        png_origins = png_data.get("info", {}).get("cornerCoordinates", {})
        pprocessor = PngProcessor(png_path,
                                  png_origins,
                                  x_dist, y_dist,
                                  quadrant)
        output_space = self.get_output_space()
        pprocessor.make_tile(output_resolution[0], output_resolution[1],
                             output_space[0], output_space[1],
                             output_path)

        # CSV出力
        params_for_csv = png_data.get("params_for_csv", {})
        self.make_csv(output_path, params_for_csv)

        QMessageBox.information(None, '処理終了', 'ファイルの出力が完了しました')
        self.dlg.show()

    def get_output_resolution(self):
        """[summary]
        UIの状態を判別して出力解像度を取得する
        Returns:
            [int, int]: [x解像度, y解像度]
        """
        output_resolution = self.dlg.outputResolutionsComboBox.currentData()
        if output_resolution is None:
            output_resolution = [self.dlg.customWidthSpinBox.value(),
                                 self.dlg.customHeightSpinBox.value()]
        return output_resolution


    def get_output_space(self):
      output_space = [self.dlg.outputSpace.value(), self.dlg.outputSpace.value()]
      return output_space

    def make_csv(self, output_dir: str, params: dict):
        csv_path = os.path.join(output_dir, "info.csv")
        with open(csv_path, mode="w", encoding='shift_jis') as f:
            writer = csv.writer(f, lineterminator="\n")
            writer.writerow(params.keys())
            writer.writerow(params.values())
